/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.ships.fetchAndIssue;

import java.util.Iterator;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.codeStorage.CodeStorage;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.TriggerAction;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * @author ib27688
 *
 */
public class FetchAndIssueTestSource extends Component {
	
	Terminal out;
	Terminal ackOut;
	
	FetchIssueTestSourceAction srcAction;
	
	public FetchAndIssueTestSource(String name, Parameters params) {
		super(name, null);
		
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
		
				director.error("Component: " + name + ", id= " + getID() 
						+ ": Attribute " + pName + " with value " + sValue
						+ " is ignored");					
		
			} // for
		} // if params != null
		
		buildFetchAndIssueTestSource();

		
	} // FifoSource
	
	protected void buildFetchAndIssueTestSource() {
		srcAction= new FetchIssueTestSourceAction("FetchIssueTestSourceAction");
		addMember(srcAction);
		out= new OutputTerminal(FleetGlobal.OUT, this);
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT, this);
		//attach the terminals
		srcAction.attach(out);
		srcAction.attach(ackOut);
				
	} // buildFetchAndIssueTestSource
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.OUT)) {
			return out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT)) {
			return ackOut;
		}
		else {
			// unknown terminal
			return null;
		}
	} // getTerminal
	
	public String stateInfo() {
		
		String onOff = srcAction.isEnabled()
			? "FetchIssueTestSource is enabled." 
			: "FetchIssueTestSource is disabled."; 
		
		return "FetchIssueTestSource " + getName() + ", id= " + getID() + ",  has generated " 
		+ srcAction.count 
		+ " items. The next data item to be generated is "
		+ srcAction.next
		+ ". " 
		+ onOff;
	} // stateInfo
	
	private class FetchIssueTestSourceAction extends TriggerAction {
		
		String next;
		int count= 0;

		// iterator through code bags
		Iterator<String> si= CodeStorage.getInstance().bagNameIterator();
		
		public FetchIssueTestSourceAction(String name) {
			super(name);
			
			// next data item to produce
			if (si.hasNext()) {
				next= si.next();
				//start enabled right away
				enabled= true;
			}
			else {
				next= null;
				enabled= false;
			}
		
			// how many data items to produce
			outputFunctions.put(FleetGlobal.OUT, new GenerateData());
			
		} // FifoSourceAction
		
		/** Generate new data items */
		private class GenerateData extends Function {

			private boolean displayGenerated;
			
			public GenerateData() {
				super("GenerateData");

				// what is the delay?
				Integer fDelay= globals.intValue(FetchAndIssueTestDefaults.SOURCE_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: FetchAndIssueTestDefaults.SOURCE_DELAY_DEF;
	
				// log as you go?
				Boolean dGen= globals.booleanValue(FetchAndIssueTestDefaults.SOURCE_DISPLAY_GENERATED);	
				displayGenerated= (dGen != null)
					? dGen
					: FetchAndIssueTestDefaults.SOURCE_DISPLAY_GENERATED_DEF;
				
			} // GenerateData
			
			public Object value() {
				String current= next;
				// is this the last item to be generated?
				if (si.hasNext()) {
					next= si.next();
				}
				else {
					enabled= false;
				}

				Payload p= new Payload();
				Data d= new Data(current);
				p.data= d;
				p.source= 7;
				p.destination= 11;
				p.repeats= 1;
				p.type= Payload.PayloadType.INSTRUCTION;
				if (displayGenerated) {
					director.info("Source " 
							+ FetchAndIssueTestSource.this.getName() 
							+ ", ID= "
							+ FetchAndIssueTestSource.this.getID()
							+ ", generated payload "
							+ p);
				}
				count++;
				
				return p;	
			} // value
		} // GenerateData


	} // SourceAction
	
	
} // FetchIssueTestSource
